#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <driver_types.h>
#include <cufft.h>
#include <cuda_runtime_api.h>

#include "cutil.h"
#include "m_pd.h"
#include "pdcuda_api.h"
#include "cuconv~.h"


/* ------------------------------- cuconv~ --------------------------------- */


static t_class *cuconv_class;

void cuda_conv(t_cuconv *x);

static void cuconv_malloc(t_cuconv *x);
static void cuconv_free(t_cuconv *x);


/*static void cuconv_bang(t_cuconv *x) {
  post("bang!");
}*/

static void cuconv_gen_report(t_cuconv *x, t_symbol *s)
{
  fprintf(stdout, "%d %f %f %f %d\n",
          x->veclength,
          (float) x->timeHostToDevice / x->cycles,
          (float) x->timeKernel / x->cycles,
          (float) x->timeDeviceToHost / x->cycles,
          x->cycles);
}

/*****************************************************************************
 *                   __
 *  _ __   ___ _ __ / _| ___  _ __ _ __ ___  
 * | '_ \ / _ \ '__| |_ / _ \| '__| '_ ` _ \
 * | |_) |  __/ |  |  _| (_) | |  | | | | | |
 * | .__/ \___|_|  |_|  \___/|_|  |_| |_| |_|
 * |_|
 ****************************************************************************/

/* this is the actual performance routine which acts on the samples.
   It's called with a single pointer "w" which is our location in the
   DSP call list.  We return a new "w" which will point to the next item
   after us.  Meanwhile, w[0] is just a pointer to dsp-perform itself
   (no use to us), w[1] and w[2] are the input and output vector locations,
   and w[3] is the number of points to calculate. */
static t_int *cuconv_perform(t_int *w)
{
  t_cuconv *x = (t_cuconv *)(w[1]);
  int n = (int)(w[2]);
  int vecsize = (int)(w[2]);
  t_float *in = (t_float *)(w[4]);
  t_float *in2 = (t_float *)(w[5]);
  t_float *out1 = (t_float *)(w[6]);
  t_float time;
  int i;

  /* verify if block size changed */
  if (x->veclength != n) {
    post("cuconv~: changing block size from %d to %d...", x->veclength, n);
    post("cuconv~: changing block overlap from %d to %d...", x->overlap, vecsize);
    post("cuconv~: system samplerate: %f.", sys_getsr());
    x->veclength = n;
    cuconv_free(x);
    cuconv_malloc(x);
    x->cycles = 0;
    x->timeHostToDevice = 0; 
    x->timeKernel = 0; 
    x->timeDeviceToHost = 0; 
    x->deltat = (float) x->veclength / (sys_getsr() * x->overlap);
  }

  /* increase cycle counter */
  (x->cycles)++;

  /*--------------------------------------------------------------------------
   * 1. copy memory from host to device...
   *------------------------------------------------------------------------*/
  CUDA_SAFE_MEASURE(
    cudaMemcpy(x->d_A, in,
      (size_t) n*sizeof(t_sample),
      cudaMemcpyHostToDevice);
    cudaMemcpy(x->d_B, in2,
      (size_t) n*sizeof(t_sample),
      cudaMemcpyHostToDevice),
    x->start, x->stop, &time, 0);
  /* ...and record measured time. */
  x->timeHostToDevice += time;
  outlet_float(x->f_timeHostToDevice, x->timeHostToDevice/x->cycles);

  /*--------------------------------------------------------------------------
   * 2. Convolution
   *------------------------------------------------------------------------*/

  /* do the convolution */
  cudaEventRecord(x->start, 0);
  //cuda_conv(x);
  
  /* ...and record measured time. */
  //CUDA_SAFE_CALL(cudaThreadSynchronize());
  cudaEventRecord(x->stop, 0);
  cudaEventSynchronize(x->stop);
  cudaEventElapsedTime(&time, x->start, x->stop);
  x->timeKernel += time;
  outlet_float(x->f_timeKernel, x->timeKernel/x->cycles);

  /*--------------------------------------------------------------------------
   * 3. copy memory from device to host...
   *------------------------------------------------------------------------*/
  CUDA_SAFE_MEASURE(cudaMemcpy(out1, x->d_C,
                               (size_t) n*sizeof(t_sample),
                               cudaMemcpyDeviceToHost),
                    x->start, x->stop, &time, 0);

  //for (i=0; i<x->veclength; i++)
  //  printf("%f ", out1[i]);
  //printf("\n\n\n");
  /* ...and record measured time. */
  //printf("in: %p\n", in);
  //printf("ou: %p\n", out1);
  //memcpy(out1, in, n*(sizeof(t_sample)));
  x->timeDeviceToHost += time;
  outlet_float(x->f_timeDeviceToHost, x->timeDeviceToHost/x->cycles);

  return (w+7);
}


/* called to start DSP.  Here we call Pd back to add our perform
   routine to a linear callback list which Pd in turn calls to grind
   out the samples. */
static void cuconv_tilde_dsp(t_cuconv *x, t_signal **sp)
{
  dsp_add(cuconv_perform,
          6,
          x,                /* (1) */
          sp[0]->s_n,       /* (2) */
          sp[0]->s_vecsize, /* (3) */
          sp[0]->s_vec,     /* (4) */
          sp[1]->s_vec,     /* (5) */
          sp[2]->s_vec);    /* (6) */
}


/*****************************************************************************
 *   _ __   _____      __
 *  | '_ \ / _ \ \ /\ / /
 *  | | | |  __/\ V  V / 
 *  |_| |_|\___| \_/\_/  
 *
 * This is called everytime an object is created or DSP is turned on.
 ****************************************************************************/
static void *cuconv_new(void)
{
  post("initializing cuconv~ object.");

  /* init pd object */
  t_cuconv *x = (t_cuconv *)pd_new(cuconv_class);
  inlet_new(&x->x_obj, &x->x_obj.ob_pd, &s_signal, &s_signal);
  outlet_new(&x->x_obj, gensym("signal"));
  x->f_timeHostToDevice = outlet_new(&x->x_obj, gensym("float"));
  x->f_timeKernel = outlet_new(&x->x_obj, gensym("float"));
  x->f_timeDeviceToHost = outlet_new(&x->x_obj, gensym("float"));
  x->veclength = 64;
  x->overlap = 1; /* TODO: allow for other overlaps */
  x->srate = sys_getsr();
  x->deltat = x->veclength / (sys_getsr() * x->overlap);
  x->cycles = 0;
  x->timeHostToDevice = 0;
  x->timeKernel = 0;
  x->timeDeviceToHost = 0;

  /* choose cuda device */
  lib_set_cuda_device(1);

  /* allocate memory */
  cuconv_malloc(x);

  CUDA_SAFE_CALL(cudaThreadSynchronize());
  return (x);
}


/*****************************************************************************
 * cuconv_malloc(void)
 *
 * This allocates memory on host and device.
 */
static void cuconv_malloc(t_cuconv *x)
{
  int i;

  /* allocate device memory */
  CUT_SAFE_MALLOC(cudaMalloc((void **) &x->d_A, x->veclength*sizeof(t_float)));
  CUT_SAFE_MALLOC(cudaMalloc((void **) &x->d_B, x->veclength*sizeof(t_float)));
  CUT_SAFE_MALLOC(cudaMalloc((void **) &x->d_C, x->veclength*sizeof(t_float)));
  x->tmp = malloc(x->veclength*sizeof(t_float));

  CUDA_SAFE_CALL(cudaThreadSynchronize());

  /* create events for timing */
  cudaEventCreate(&(x->start));
  cudaEventCreate(&(x->stop));

}


/*****************************************************************************
 * cuconv_free(void)
 *
 * This is called everytime an object is deleted or DSP is stopped.
 */
static void cuconv_free(t_cuconv *x)
{
  /* free host memory */
  /*free(x->h_A);*/
  /* free device memory */
  cudaFree((void *) x->d_A);
  cudaFree((void *) x->d_B);
  cudaFree((void *) x->d_C);
  /* destroy events for timing */
  cudaEventDestroy(x->start);
  cudaEventDestroy(x->stop);
}


/*****************************************************************************
 * cuconv_tilde_setup(void)
 *
 * this routine, which must have exactly this name (with the "~" replaced
 * by "_tilde) is called when the code is first loaded, and tells Pd how
 * to build the "class".
 */
void cuconv_tilde_setup(void)
{
  post("inicializando classe cuconv~...");
  cuconv_class = class_new(gensym("cuconv~"), (t_newmethod)cuconv_new, 0,
                                  sizeof(t_cuconv), 0, A_DEFFLOAT, 0);
  /* this is magic to declare that the leftmost, "main" inlet
     takes signals; other signal inlets are done differently...*/
  CLASS_MAINSIGNALIN(cuconv_class, t_cuconv, x_f);
  /* here we tell Pd about the "dsp" method, which is called back
     when DSP is turned on. */
  class_addmethod(cuconv_class, (t_method) cuconv_tilde_dsp, gensym("dsp"), (t_atomtype) 0);
  class_addmethod(cuconv_class, (t_method) cuconv_gen_report, gensym("gen-report"), (t_atomtype) 0);

}

