#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <cuda_runtime_api.h>

#include "m_pd.h"
#include "pdcuda_api.h"


/* ------------------------ cuda~ ----------------------------- */

static t_class *cuda_class;

typedef struct _cuda
{
  t_object x_obj;       /* obligatory header */
  t_float x_f;          /* place to hold inlet's value if it's set by message */
  unsigned int cycles;
  float factor1;
  float factor2;
  float timeHostToDevice; 
  float timeDeviceToDevice; 
  float timeKernel; 
  float timeDeviceToHost; 
  t_outlet *f_timeHostToDevice; 
  t_outlet *f_timeDeviceToDevice; 
  t_outlet *f_timeKernel; 
  t_outlet *f_timeDeviceToHost; 
  cudaEvent_t start, stop;
  float *h_A, *h_B; /* pointers for memory on host */
  float *d_A, *d_B; /* pointers for memory on device */
} t_cuda;


/*static void cuda_bang(t_cuda *x) {
  post("bang!");
}

static void cuda_anything(t_cuda *x, t_symbol *s, int argc, t_atom *argv)
{
}*/

/* this is the actual performance routine which acts on the samples.
   It's called with a single pointer "w" which is our location in the
   DSP call list.  We return a new "w" which will point to the next item
   after us.  Meanwhile, w[0] is just a pointer to dsp-perform itself
   (no use to us), w[1] and w[2] are the input and output vector locations,
   and w[3] is the number of points to calculate. */
static t_int *cuda_perform(t_int *w)
{
  t_cuda *x = (t_cuda *)(w[1]);
  int n = (int)(w[2]);
  t_float *in = (t_float *)(w[3]);
  t_float *out1 = (t_float *)(w[4]);
  t_float *out2 = (t_float *)(w[5]);
  float time;
  cudaError_t errno;

  /* increase cycle counter */
  (x->cycles)++;

  /* 1. copy memory from host to device... */
  cudaEventRecord(x->start, 0);
  errno = cudaMemcpy(x->d_A, in, (size_t) n*sizeof(t_sample), cudaMemcpyHostToDevice);
  if (errno)
    printf("cuda error (perform 1): %d - %s\n", (int) errno, cudaGetErrorString(errno));
  /* ... and measure time. */
  cudaEventRecord(x->stop, 0);
  cudaEventSynchronize(x->stop);
  cudaEventElapsedTime(&time, x->start, x->stop);
  x->timeHostToDevice += time;
  outlet_float(x->f_timeHostToDevice, x->timeHostToDevice/x->cycles);

  /* 2. copy memory from device to device... */
  cudaEventRecord(x->start, 0);
  errno = cudaMemcpy((void *) x->d_A + n*sizeof(t_sample),
                     x->d_A,
                     (size_t) n*sizeof(t_sample),
                     cudaMemcpyDeviceToDevice);
  if (errno)
    printf("cuda error (perform 2): %d - %s\n", (int) errno, cudaGetErrorString(errno));
  /* ...and measure time. */
  cudaEventRecord(x->stop, 0);
  cudaEventSynchronize(x->stop);
  cudaEventElapsedTime(&time, x->start, x->stop);
  x->timeDeviceToDevice += time;
  outlet_float(x->f_timeDeviceToDevice, x->timeDeviceToDevice/x->cycles);

  /* 3. run kernels... */
  cudaSetDeviceFlags(cudaDeviceBlockingSync);
  cudaEventRecord(x->start, 0);
  cuda_scalarplus_basic(x->d_A, x->d_A, x->factor1, n); /* <--- kernel is called here */
  cuda_scalarplus_basic((void *) x->d_A + n*sizeof(t_sample),
                        (void *) x->d_A + n*sizeof(t_sample),
                        x->factor2, n); /* <--- kernel is called here */
  /* ...and measure time. */
  cudaEventRecord(x->stop, 0);
  cudaEventSynchronize(x->stop);
  cudaEventElapsedTime(&time, x->start, x->stop);
  x->timeKernel += time;
  outlet_float(x->f_timeKernel, x->timeKernel/x->cycles);

  /* 4. copy memory from device to host... */
  cudaEventRecord(x->start, 0);
  errno = cudaMemcpy(x->h_A,
                     x->d_A,
                     (size_t) 2*n*sizeof(t_sample),
                     cudaMemcpyDeviceToHost);
  if (errno)
    printf("cuda error (perform 3): %d - %s\n", (int) errno, cudaGetErrorString(errno));
  /* ...and measure time. */
  cudaEventRecord(x->stop, 0);
  cudaEventSynchronize(x->stop);
  cudaEventElapsedTime(&time, x->start, x->stop);
  x->timeDeviceToHost += time;
  outlet_float(x->f_timeDeviceToHost, x->timeDeviceToHost/x->cycles);

  /* copy signal to outlets */
  memcpy(out1, x->h_A, sizeof(t_float) * n);
  memcpy(out2, (void *) x->h_A + n*sizeof(t_sample), sizeof(t_float) * n);

  return (w+6);
}


/* called to start DSP.  Here we call Pd back to add our perform
   routine to a linear callback list which Pd in turn calls to grind
   out the samples. */
static void cuda_tilde_dsp(t_cuda *x, t_signal **sp)
{
  dsp_add(cuda_perform, 5, x, sp[0]->s_n, sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec);
}


/**
 * cuda_new(void)
 *
 * This is called everytime an object is created or DSP is turned on.
 */
static void *cuda_new(void)
{
  int N = 64;
  cudaError_t errno;
  
  post("initializing cuda~ object.");

  /* init pd object */
  t_cuda *x = (t_cuda *)pd_new(cuda_class);
  floatinlet_new(&x->x_obj, &x->factor1);
  floatinlet_new(&x->x_obj, &x->factor2);
  outlet_new(&x->x_obj, gensym("signal"));
  outlet_new(&x->x_obj, gensym("signal"));
  x->f_timeHostToDevice = outlet_new(&x->x_obj, gensym("float"));
  x->f_timeDeviceToDevice = outlet_new(&x->x_obj, gensym("float"));
  x->f_timeKernel = outlet_new(&x->x_obj, gensym("float"));
  x->f_timeDeviceToHost = outlet_new(&x->x_obj, gensym("float"));
  x->cycles = 0;
  x->factor1 = 0;
  x->factor2 = 0;
  x->timeHostToDevice = 0;
  x->timeDeviceToDevice = 0;
  x->timeKernel = 0;
  x->timeDeviceToHost = 0;

  /* choose cuda device */
  lib_set_cuda_device(1);

  /* allocate host memory */
  errno = cudaMallocHost((void **) &x->h_A, 2*N*sizeof(t_sample));
  if (errno)
    printf("cudaMallocHost error (new): %d - %s\n", (int) errno, cudaGetErrorString(errno));
  x->h_B = x->h_A + N*sizeof(t_sample);

 /* allocate device memory */
  errno = cudaMalloc((void **) &x->d_A, 2*N*sizeof(t_sample));
  if (errno)
    printf("cudaMalloc error (new): %d - %s\n", (int) errno, cudaGetErrorString(errno));
  /*errno = cudaMalloc((void **) &x->d_B, N*sizeof(t_sample));
  if (errno)
    printf("cudaMalloc error (new): %d - %s\n", (int) errno,
    cudaGetErrorString(errno));*/
  x->d_B = x->d_A + N*sizeof(t_sample);

  /* create events for timing */
  cudaEventCreate(&(x->start));
  cudaEventCreate(&(x->stop));

  return (x);
}


/**
 * cuda_free(void)
 *
 * This is called everytime an object is deleted or DSP is stopped.
 */
void cuda_free(t_cuda *x)
{
  cudaError_t errno;
  /* free host memory */
  free(x->h_A);
  /* free device memory */
  errno = cudaFree((void **) &x->d_A);
  if (errno)
    printf("cuda error (free): %d - %s\n", (int) errno, cudaGetErrorString(errno));
  /* destroy events for timing */
  cudaEventDestroy(x->start);
  cudaEventDestroy(x->stop);
}


/**
 * cuda_tilde_setup(void)
 *
 * this routine, which must have exactly this name (with the "~" replaced
 * by "_tilde) is called when the code is first loaded, and tells Pd how
 * to build the "class".
 */
void cuda_tilde_setup(void)
{
  post("inicializando classe cuda~...");
  cuda_class = class_new(gensym("cuda~"), (t_newmethod)cuda_new, 0,
                                  sizeof(t_cuda), 0, A_DEFFLOAT, 0);
  /* this is magic to declare that the leftmost, "main" inlet
     takes signals; other signal inlets are done differently...*/
  CLASS_MAINSIGNALIN(cuda_class, t_cuda, x_f);
  /* here we tell Pd about the "dsp" method, which is called back
     when DSP is turned on. */
  class_addmethod(cuda_class, (t_method) cuda_tilde_dsp, gensym("dsp"), (t_atomtype) 0);
  /*class_addbang(cuda_class, (t_method)cuda_bang);
  class_addanything(cuda_class, cuda_anything);*/
}
