#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <driver_types.h>
#include <cufft.h>
#include <cuda_runtime_api.h>

#include "cutil.h"
#include "m_pd.h"
#include "pdcuda_api.h"


/* ------------------------------- cufft~ --------------------------------- */

void cuda_logmagnitude2(cufftComplex *c, cufftReal *a, int n);

static t_class *cufft_class;

typedef struct _cufft
{
  t_object x_obj;       /* obligatory header */
  t_float x_f;          /* place to hold inlet's value if it's set by message */
  int veclength;
  unsigned int cycles;
  float factor1;
  float factor2;
  float timeHostToDevice; 
  float timeKernel; 
  float timeDeviceToHost; 
  t_outlet *f_timeHostToDevice; 
  t_outlet *f_timeKernel; 
  t_outlet *f_timeDeviceToHost; 
  t_outlet *f_setBlock;
  t_outlet *s_switchDsp; 
  cudaEvent_t start, stop;
  /* pointers for memory on host */
  /*cufftReal *h_A;*/ 
  /* pointers for memory on device */
  cufftReal *d_A;
  cufftComplex *d_B;
  /* CUFFT */
  cufftHandle plan;
} t_cufft;

static void cufft_malloc(t_cufft *x);
static void cufft_free(t_cufft *x);


/*static void cufft_bang(t_cufft *x) {
  post("bang!");
}*/

static void cufft_gen_report(t_cufft *x, t_symbol *s)
{
  fprintf(stdout, "%d %f %f %f\n",
          x->veclength,
          (float) x->timeHostToDevice / x->cycles,
          (float) x->timeKernel / x->cycles,
          (float) x->timeDeviceToHost / x->cycles);
}

/* this is the actual performance routine which acts on the samples.
   It's called with a single pointer "w" which is our location in the
   DSP call list.  We return a new "w" which will point to the next item
   after us.  Meanwhile, w[0] is just a pointer to dsp-perform itself
   (no use to us), w[1] and w[2] are the input and output vector locations,
   and w[3] is the number of points to calculate. */
static t_int *cufft_perform(t_int *w)
{
  t_cufft *x = (t_cufft *)(w[1]);
  int n = (int)(w[2]);
  t_float *in = (t_float *)(w[3]);
  t_float *out1 = (t_float *)(w[4]);
  float time;

  /* verify if block size changed */
  if (x->veclength != n) {
    post("cufft~: changing block size from %d to %d...", x->veclength, n);
    x->veclength = n;
    cufft_free(x);
    cufft_malloc(x);
    x->cycles = 0;
    x->timeHostToDevice = 0; 
    x->timeKernel = 0; 
    x->timeDeviceToHost = 0; 
  }

  /* increase cycle counter */
  (x->cycles)++;

  /* 1. copy memory from host to device... */
  CUDA_SAFE_MEASURE(cudaMemcpy(x->d_A, in,
                               (size_t) n*sizeof(t_sample),
                               cudaMemcpyHostToDevice),
                    x->start, x->stop, &time, 0);
  /* ...and record measured time. */
  x->timeHostToDevice += time;
  outlet_float(x->f_timeHostToDevice, x->timeHostToDevice/x->cycles);

  /* 2. execute FFT on signal and extract magnitude... */
  CUDA_SAFE_MEASURE(cufftExecR2C(x->plan, x->d_A, x->d_B), /* <- kernel call */
                    x->start, x->stop, &time, 0);
  CUDA_SAFE_CALL(cudaThreadSynchronize());
  cuda_logmagnitude2(x->d_B, x->d_A, n/2+1); /* <- kernel call */
  /* ...and record measured time. */
  x->timeKernel += time;
  outlet_float(x->f_timeKernel, x->timeKernel/x->cycles);

  /* 3. copy memory from device to host... */
  CUDA_SAFE_MEASURE(cudaMemcpy(out1, x->d_A,
                               (size_t) n*sizeof(t_sample),
                               cudaMemcpyDeviceToHost),
                    x->start, x->stop, &time, 0);
  /* ...and record measured time. */
  x->timeDeviceToHost += time;
  outlet_float(x->f_timeDeviceToHost, x->timeDeviceToHost/x->cycles);

  return (w+5);
}


/* called to start DSP.  Here we call Pd back to add our perform
   routine to a linear callback list which Pd in turn calls to grind
   out the samples. */
static void cufft_tilde_dsp(t_cufft *x, t_signal **sp)
{
  dsp_add(cufft_perform, 4, x, sp[0]->s_n, sp[0]->s_vec, sp[1]->s_vec);
}


/**
 * cufft_new(void)
 *
 * This is called everytime an object is created or DSP is turned on.
 */
static void *cufft_new(void)
{
  post("initializing cufft~ object.");

  /* init pd object */
  t_cufft *x = (t_cufft *)pd_new(cufft_class);
  outlet_new(&x->x_obj, gensym("signal"));
  x->f_timeHostToDevice = outlet_new(&x->x_obj, gensym("float"));
  x->f_timeKernel = outlet_new(&x->x_obj, gensym("float"));
  x->f_timeDeviceToHost = outlet_new(&x->x_obj, gensym("float"));
  x->f_setBlock = outlet_new(&x->x_obj, gensym("symbol"));
  x->s_switchDsp = outlet_new(&x->x_obj, gensym("symbol"));
  x->veclength = 64;
  x->cycles = 0;
  x->timeHostToDevice = 0;
  x->timeKernel = 0;
  x->timeDeviceToHost = 0;

  /* choose cuda device */
  lib_set_cuda_device(1);

  /* allocate memory */
  cufft_malloc(x);

  return (x);
}


/**
 * cufft_malloc(void)
 *
 * This allocates memory on host and device.
 */
static void cufft_malloc(t_cufft *x)
{
  /* allocate host memory */
  /*CUT_SAFE_MALLOC(cudaMallocHost((void **) &x->h_A, 2*x->veclength*sizeof(cufftReal)));*/

  /* allocate device memory */
  CUT_SAFE_MALLOC(cudaMalloc((void **) &x->d_A, x->veclength*sizeof(cufftReal)));
  CUT_SAFE_MALLOC(cudaMalloc((void **) &x->d_B, x->veclength*sizeof(cufftComplex)/2+1));

  /* create events for timing */
  cudaEventCreate(&(x->start));
  cudaEventCreate(&(x->stop));

  /* CUFFT plan */
  CUFFT_SAFE_CALL(cufftPlan1d(&x->plan, x->veclength, CUFFT_R2C, 1));
}


/**
 * cufft_free(void)
 *
 * This is called everytime an object is deleted or DSP is stopped.
 */
static void cufft_free(t_cufft *x)
{
  /* free host memory */
  /*free(x->h_A);*/
  /* free device memory */
  cudaFree((void *) x->d_A);
  cudaFree((void *) x->d_B);
  /* destroy events for timing */
  cudaEventDestroy(x->start);
  cudaEventDestroy(x->stop);
  /* cufft */
  cufftDestroy(x->plan);
}


/**
 * cufft_tilde_setup(void)
 *
 * this routine, which must have exactly this name (with the "~" replaced
 * by "_tilde) is called when the code is first loaded, and tells Pd how
 * to build the "class".
 */
void cufft_tilde_setup(void)
{
  post("inicializando classe cufft~...");
  cufft_class = class_new(gensym("cufft~"), (t_newmethod)cufft_new, 0,
                                  sizeof(t_cufft), 0, A_DEFFLOAT, 0);
  /* this is magic to declare that the leftmost, "main" inlet
     takes signals; other signal inlets are done differently...*/
  CLASS_MAINSIGNALIN(cufft_class, t_cufft, x_f);
  /* here we tell Pd about the "dsp" method, which is called back
     when DSP is turned on. */
  class_addmethod(cufft_class, (t_method) cufft_tilde_dsp, gensym("dsp"), (t_atomtype) 0);
  class_addmethod(cufft_class, (t_method) cufft_gen_report, gensym("gen-report"), (t_atomtype) 0);
}
