#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <driver_types.h>
#include <cufft.h>
#include <cuda_runtime_api.h>

#include "cutil.h"
#include "m_pd.h"
#include "pdcuda_api.h"
#include "cupv~.h"


/* ------------------------------- cupv~ --------------------------------- */


static t_class *cupv_class;

void cuda_pv(t_cupv *x);
void cuda_pv_set_texture(t_cupv *x);

static void cupv_malloc(t_cupv *x);
static void cupv_free(t_cupv *x);


/*static void cupv_bang(t_cupv *x) {
  post("bang!");
}*/

static void cupv_gen_report(t_cupv *x, t_symbol *s)
{
  fprintf(stdout, "%d %f %f %f %d\n",
          x->veclength,
          (float) x->timeHostToDevice / x->cycles,
          (float) x->timeKernel / x->cycles,
          (float) x->timeDeviceToHost / x->cycles,
          x->cycles);
}

/*****************************************************************************
 *                   __
 *  _ __   ___ _ __ / _| ___  _ __ _ __ ___  
 * | '_ \ / _ \ '__| |_ / _ \| '__| '_ ` _ \
 * | |_) |  __/ |  |  _| (_) | |  | | | | | |
 * | .__/ \___|_|  |_|  \___/|_|  |_| |_| |_|
 * |_|
 ****************************************************************************/

/* this is the actual performance routine which acts on the samples.
   It's called with a single pointer "w" which is our location in the
   DSP call list.  We return a new "w" which will point to the next item
   after us.  Meanwhile, w[0] is just a pointer to dsp-perform itself
   (no use to us), w[1] and w[2] are the input and output vector locations,
   and w[3] is the number of points to calculate. */
static t_int *cupv_perform(t_int *w)
{
  t_cupv *x = (t_cupv *)(w[1]);
  int n = (int)(w[2]);
  int vecsize = (int)(w[2]);
  t_float *in = (t_float *)(w[4]);
  t_float *out1 = (t_float *)(w[5]);
  t_float time;
  int i;

  /* verify if block size changed */
  if (x->veclength != n) {
    post("cupv~: changing block size from %d to %d...", x->veclength, n);
    post("cupv~: changing block overlap from %d to %d...", x->overlap, vecsize);
    post("cupv~: system samplerate: %f.", sys_getsr());
    x->veclength = n;
    cupv_free(x);
    cupv_malloc(x);
    x->cycles = 0;
    x->timeHostToDevice = 0; 
    x->timeKernel = 0; 
    x->timeDeviceToHost = 0; 
    x->deltat = (float) x->veclength / (sys_getsr() * x->overlap);
  }

  /* increase cycle counter */
  (x->cycles)++;

  /*--------------------------------------------------------------------------
   * 1. copy memory from host to device...
   *------------------------------------------------------------------------*/
  CUDA_SAFE_MEASURE(cudaMemcpy(x->d_A, in,
                               (size_t) n*sizeof(t_sample),
                               cudaMemcpyHostToDevice),
                    x->start, x->stop, &time, 0);
  /* ...and record measured time. */
  x->timeHostToDevice += time;
  outlet_float(x->f_timeHostToDevice, x->timeHostToDevice/x->cycles);

  /*--------------------------------------------------------------------------
   * 2. Phase Vocoder
   *------------------------------------------------------------------------*/
  /* FFT */

  /* do the vocoder */
  cudaEventRecord(x->start, 0);
  cuda_pv(x);
  
  /* ...and record measured time. */
  //CUDA_SAFE_CALL(cudaThreadSynchronize());
  cudaEventRecord(x->stop, 0);
  cudaEventSynchronize(x->stop);
  cudaEventElapsedTime(&time, x->start, x->stop);
  x->timeKernel += time;
  outlet_float(x->f_timeKernel, x->timeKernel/x->cycles);

  /*--------------------------------------------------------------------------
   * 3. copy memory from device to host...
   *------------------------------------------------------------------------*/
  CUDA_SAFE_MEASURE(cudaMemcpy(out1, x->d_H,
                               (size_t) n*sizeof(t_sample),
                               cudaMemcpyDeviceToHost),
                    x->start, x->stop, &time, 0);
  //CUDA_SAFE_CALL(cudaMemcpy(x->tmp, x->d_A,
  //                          (size_t) n*sizeof(t_sample),
  //                          cudaMemcpyDeviceToHost));

  //for (i=0; i<x->veclength; i++)
  //  printf("%f ", out1[i]);
  //printf("\n\n\n");
  /* ...and record measured time. */
  //printf("in: %p\n", in);
  //printf("ou: %p\n", out1);
  //memcpy(out1, in, n*(sizeof(t_sample)));
  x->timeDeviceToHost += time;
  outlet_float(x->f_timeDeviceToHost, x->timeDeviceToHost/x->cycles);

  return (w+6);
}


/* called to start DSP.  Here we call Pd back to add our perform
   routine to a linear callback list which Pd in turn calls to grind
   out the samples. */
static void cupv_tilde_dsp(t_cupv *x, t_signal **sp)
{
  dsp_add(cupv_perform,
          5,
          x,                /* (1) */
          sp[0]->s_n,       /* (2) */
          sp[0]->s_vecsize, /* (3) */
          sp[0]->s_vec,     /* (4) */
          sp[1]->s_vec);    /* (5) */
}


/*****************************************************************************
 *   _ __   _____      __
 *  | '_ \ / _ \ \ /\ / /
 *  | | | |  __/\ V  V / 
 *  |_| |_|\___| \_/\_/  
 *
 * This is called everytime an object is created or DSP is turned on.
 ****************************************************************************/
static void *cupv_new(void)
{
  post("initializing cupv~ object.");

  /* init pd object */
  t_cupv *x = (t_cupv *)pd_new(cupv_class);
  outlet_new(&x->x_obj, gensym("signal"));
  x->f_timeHostToDevice = outlet_new(&x->x_obj, gensym("float"));
  x->f_timeKernel = outlet_new(&x->x_obj, gensym("float"));
  x->f_timeDeviceToHost = outlet_new(&x->x_obj, gensym("float"));
  x->veclength = 64;
  x->overlap = 1; /* TODO: allow for other overlaps */
  x->srate = sys_getsr();
  x->deltat = x->veclength / (sys_getsr() * x->overlap);
  x->cycles = 0;
  x->timeHostToDevice = 0;
  x->timeKernel = 0;
  x->timeDeviceToHost = 0;
  x->s = 1024; /* sine table size */

  /* inlets and outlets */
  floatinlet_new (&x->x_obj , &x->alpha) ;
  floatinlet_new (&x->x_obj , &x->beta) ;

  /* choose cuda device */
  lib_set_cuda_device(0);

  /* allocate memory */
  cupv_malloc(x);
  
  /* zero out auxiliary vectors */
  cuda_zero_basic(x->d_C, x->veclength);
  cuda_zero_basic(x->d_D, x->veclength);
  cuda_zero_basic(x->d_E, x->veclength);
  cuda_zero_basic(x->d_F, x->veclength);
  cuda_zero_basic(x->d_G, x->veclength);
  cuda_zero_basic(x->d_H, x->veclength);

  /* textures*/
  cuda_pv_set_texture(x);

  CUDA_SAFE_CALL(cudaThreadSynchronize());
  return (x);
}


/*****************************************************************************
 * cupv_malloc(void)
 *
 * This allocates memory on host and device.
 */
static void cupv_malloc(t_cupv *x)
{
  int i;
  /* allocate host memory */
  /*CUT_SAFE_MALLOC(cudaMallocHost((void **) &x->h_A, 2*x->veclength*sizeof(cufftReal)));*/
  x->sineTextureFormatDesc = cudaCreateChannelDesc(32, 0, 0, 0,
                                                   cudaChannelFormatKindFloat);


  /* allocate device memory */
  CUT_SAFE_MALLOC(cudaMalloc((void **) &x->d_A, x->veclength*sizeof(cufftReal)+2));
  CUT_SAFE_MALLOC(cudaMalloc((void **) &x->d_B, x->veclength*sizeof(cufftComplex)/2+1));
  CUT_SAFE_MALLOC(cudaMalloc((void **) &x->d_C, x->veclength*sizeof(t_float)));
  CUT_SAFE_MALLOC(cudaMalloc((void **) &x->d_D, x->veclength*sizeof(t_float)));
  CUT_SAFE_MALLOC(cudaMalloc((void **) &x->d_E, x->veclength*sizeof(t_float)));
  CUT_SAFE_MALLOC(cudaMalloc((void **) &x->d_F, x->veclength*sizeof(t_float)));
  CUT_SAFE_MALLOC(cudaMalloc((void **) &x->d_G, x->veclength*sizeof(t_float)));
  CUT_SAFE_MALLOC(cudaMalloc((void **) &x->d_H, x->veclength*sizeof(t_float)));
  CUT_SAFE_MALLOC(cudaMalloc((void **) &x->d_sine, x->s*sizeof(t_float)));
  x->tmp = malloc(x->veclength*sizeof(t_float));

  /* sine table */
  x->h_sine = (t_float*)malloc(x->s*sizeof(t_float));
  for (i=0;i<x->s;i++)
    x->h_sine[i] = sin(2*M_PI*i/x->s);

  CUDA_SAFE_CALL(cudaThreadSynchronize());
  CUDA_SAFE_CALL(cudaMemcpy(x->d_sine,
                            x->h_sine,
                            (size_t) (x->s)*sizeof(t_float),
                            cudaMemcpyHostToDevice));
  /* sine texture array */
  CUDA_SAFE_CALL(cudaMallocArray(&(x->sineTextureArray),
                                 &(x->sineTextureFormatDesc), x->s, 1, 0));
  CUDA_SAFE_CALL(cudaMemcpyToArray(x->sineTextureArray,
                                   0,
                                   0,
                                   x->h_sine,
                                   (size_t) (x->s)*sizeof(t_float),
                                   cudaMemcpyHostToDevice));
  CUDA_SAFE_CALL(cudaThreadSynchronize());

  /* create events for timing */
  cudaEventCreate(&(x->start));
  cudaEventCreate(&(x->stop));

  /* CUFFT plan */
  CUFFT_SAFE_CALL(cufftPlan1d(&x->plan, x->veclength, CUFFT_R2C, 1));

  /* textures */
}


/*****************************************************************************
 * cupv_free(void)
 *
 * This is called everytime an object is deleted or DSP is stopped.
 */
static void cupv_free(t_cupv *x)
{
  /* free host memory */
  /*free(x->h_A);*/
  /* free device memory */
  cudaFree((void *) x->d_A);
  cudaFree((void *) x->d_B);
  cudaFree((void *) x->d_C);
  cudaFree((void *) x->d_D);
  cudaFree((void *) x->d_E);
  cudaFree((void *) x->d_F);
  cudaFree((void *) x->d_G);
  cudaFree((void *) x->d_H);
  /* destroy events for timing */
  cudaEventDestroy(x->start);
  cudaEventDestroy(x->stop);
  /* cufft */
  cufftDestroy(x->plan);
  /* textures */
  //free(x->sineTable);
  //CUDA_SAFE_CALL(cudaUnbindTexture(x->sineTextureRef));
}


/*****************************************************************************
 * cupv_tilde_setup(void)
 *
 * this routine, which must have exactly this name (with the "~" replaced
 * by "_tilde) is called when the code is first loaded, and tells Pd how
 * to build the "class".
 */
void cupv_tilde_setup(void)
{
  post("inicializando classe cupv~...");
  cupv_class = class_new(gensym("cupv~"), (t_newmethod)cupv_new, 0,
                                  sizeof(t_cupv), 0, A_DEFFLOAT, 0);
  /* this is magic to declare that the leftmost, "main" inlet
     takes signals; other signal inlets are done differently...*/
  CLASS_MAINSIGNALIN(cupv_class, t_cupv, x_f);
  /* here we tell Pd about the "dsp" method, which is called back
     when DSP is turned on. */
  class_addmethod(cupv_class, (t_method) cupv_tilde_dsp, gensym("dsp"), (t_atomtype) 0);
  class_addmethod(cupv_class, (t_method) cupv_gen_report, gensym("gen-report"), (t_atomtype) 0);

}

