#include <stdio.h>

#include <cuda_runtime_api.h>
#include <cufft.h>

#include "cutil.h"
#include "m_pd.h"
#include "pdcuda_api.h"
#include "cupv~.h"

extern "C" void get_grid_and_block(int *grid,int *block,int s_n);

texture<float, cudaTextureType1D, cudaReadModeElementType> sineTexture;

/*---------------------------- Function Headers ----------------------------*/


__device__ t_float interp4(t_float *T, t_float i, int N);
__device__ t_float interpl(t_float *T, t_float i, int N);
__device__ t_float modulo2pi(t_float x);
__device__ t_float modS(t_float y, int S);

/*---------------------------- Device Functions ----------------------------*/


/*****************************************************************************
 *     _    _   _    _    _  __   ______ ___ ____  
 *    / \  | \ | |  / \  | | \ \ / / ___|_ _/ ___| 
 *   / _ \ |  \| | / _ \ | |  \ V /\___ \| |\___ \
 *  / ___ \| |\  |/ ___ \| |___| |  ___) | | ___) |
 * /_/   \_\_| \_/_/   \_\_____|_| |____/___|____/ 
 ****************************************************************************/
__global__ void PV_Analysis(cufftComplex *x,
                            t_float *mag,
                            t_float *last_mag,
                            t_float *phase,
                            t_float *last_phase,
                            int n,
                            int overlap,
                            t_float srate,
                            t_float beta,
                            int s,
                            cufftHandle plan)
{
  int k = blockDim.x * blockIdx.x + threadIdx.x;
  t_float ph, deltaphi;
  /* extract magnitudes */
  mag[k] = 2*sqrtf(powf(x[k].x,2)+powf(x[k].y,2))/n;
  /* and frequency estimate */
  ph = atan2(x[k].y, x[k].x);
  deltaphi = modulo2pi(overlap * (ph - last_phase[k]));
  phase[k] = (deltaphi / (2*M_PI) + k) * srate / n;
  last_phase[k] = ph;
}


/*****************************************************************************
 *  ______   ___   _ _____ _   _ _____ ____ ___ ____  
 * / ___\ \ / / \ | |_   _| | | | ____/ ___|_ _/ ___|
 * \___ \\ V /|  \| | | | | |_| |  _| \___ \| |\___ \
 *  ___) || | | |\  | | | |  _  | |___ ___) | | ___) |
 * |____/ |_| |_| \_| |_| |_| |_|_____|____/___|____/
 ****************************************************************************/
__global__ void PV_Synthesis(t_float *mag,
                             t_float *last_mag,
                             t_float *deltak,
                             t_float *ind,
                             t_float *out,
                             t_float *phase,
                             int n,
                             int overlap,
                             t_float alpha,
                             t_float beta,
                             t_float *sine,
                             int s,
                             int srate,
                             struct textureReference * sineTextureRef)
{
  int k, t = blockDim.x * blockIdx.x + threadIdx.x;
  t_float amplitude, lambda;

  /* TODO: inserir o alpha */
  lambda = (t_float) t / ((int) (n/(overlap*alpha)+0.5));
  out[t] = 0;
  for (k=1; k<n/2; k++) {
    amplitude = (1.0-lambda)*last_mag[k] + lambda*mag[k];

    /* cubic interpolation */
    //out[t] += amplitude*interp4(sine,ind[k]+t*deltak[k],s);
    
    /* linear interpolation */
    //out[t] += amplitude*interpl(sine,ind[k]+t*deltak[k],s);
    
    /* round */
    //out[t] += amplitude*sine[(int)modS(ind[k]+t*deltak[k],s)];
    
    /* builtin sine */
    //out[t] += amplitude*sin(ind[k]+t*beta*phase[k]*2*M_PI/srate);

    /* builtin texture */
    out[t] += amplitude*tex1D(sineTexture, modS(ind[k]+t*deltak[k],s)); 

    /* no calculation */
    //out[t] += 0.01;
  }
}


/*****************************************************************************
 * CalcDeltaK
 ****************************************************************************/
__global__ void CalcDeltaK(t_float *phase,
                           t_float *deltak,
                           int n,
                           t_float beta,
                           t_float srate,
                           t_float s)
{
  int k = blockDim.x * blockIdx.x + threadIdx.x;
  if (0 < k && k < n/2) {
    deltak[k] = beta*phase[k]*s/srate;
    //deltak[k] = beta*phase[k];
  }
}


/*****************************************************************************
 * PostUpdate
 ****************************************************************************/
__global__ void PostUpdate(t_float *mag,
                          t_float *last_mag,
                          t_float *deltak,
                          t_float *ind,
                          t_float *phase,
                          int n,
                          int overlap,
                          t_float alpha,
                          t_float srate,
                          t_float s)
{
  int k = blockDim.x * blockIdx.x + threadIdx.x;
  //ind[k] = modS(ind[k]+((int)(n/(overlap*alpha)+0.5))*deltak[k],s);
  //ind[k] = modS(ind[k]+((int)((n/overlap)/alpha+0.5))*deltak[k],s);
  //ind[k] = modS(ind[k]+((t_float)n/overlap)*deltak[k],s);
  //ind[k] = modulo2pi(ind[k]+n*phase[k]*2*M_PI/srate);
  ind[k] = modS(ind[k]+((t_float)n/overlap)*deltak[k],s);
  last_mag[k] = mag[k];
}


/*----------------------------- Host Functions -----------------------------*/


/*****************************************************************************
 * cuda_pv: called from cupv_perform.
 ****************************************************************************/
extern "C" void cuda_pv(t_cupv *x)
{
  int grid, block, i;
  cufftComplex *X          = x->d_B;
  t_float      *mag        = x->d_A;
  t_float      *last_phase = x->d_C;
  t_float      *last_mag   = x->d_D;
  t_float      *deltak     = x->d_E;
  t_float      *ind        = x->d_F;
  t_float      *phase      = x->d_G;
  t_float      *out        = x->d_H;
  int           n          = x->veclength;
  int           overlap    = x->overlap;
  t_float       srate      = x->srate;
  t_float       alpha      = x->alpha;
  t_float       beta       = x->beta;
  t_float      *sine       = x->d_sine;
  int           s          = x->s;
  cufftHandle   plan       = x->plan;

  get_grid_and_block(&grid, &block, n);

  /* FFT on input signal */
  CUFFT_SAFE_CALL(cufftExecR2C(plan, (cufftReal *) mag, X)); /* <- kernel call */
  /* sync */ CUDA_SAFE_CALL(cudaThreadSynchronize());
  PV_Analysis<<<grid,block>>>(X, mag, last_mag, phase, last_phase, n, overlap,
                              srate, beta, s, plan);
  /* sync */ CUDA_SAFE_CALL(cudaThreadSynchronize());
  CalcDeltaK<<<grid,block>>>(phase, deltak, n, beta, srate, s);
  /* sync */ CUDA_SAFE_CALL(cudaThreadSynchronize());
  PV_Synthesis<<<grid,block>>>(mag, last_mag, deltak, ind, out, phase, n,
                               overlap, alpha, beta, sine, s, srate,
                               x->sineTextureRef);
  /* sync */ CUDA_SAFE_CALL(cudaThreadSynchronize());
  PostUpdate<<<grid,block>>>(mag, last_mag, deltak, ind, phase, n, overlap,
                             alpha, srate, s);
  /* sync */ CUDA_SAFE_CALL(cudaThreadSynchronize());

}


/*-------------------------------- Utilities -------------------------------*/


/*****************************************************************************
 * modulo2pi
 ****************************************************************************/
__device__ t_float modulo2pi(t_float x)
{
  if (x >= 0) {
    x -= floorf(x/(2*M_PI))*2*M_PI;
    if (x > M_PI)
      x -= 2*M_PI;
  }
  else {
    x += (floorf(-x/(2*M_PI)))*2*M_PI;
    if (x < - M_PI)
      x += 2*M_PI;
  }
  return x;
}


/*****************************************************************************
 * interp4
 ****************************************************************************/
__device__ t_float interp4(t_float *T, t_float i, int N)
{
  t_float y = i-(int)i;
  int i0,i1,i2,i3;

  i1 = ((int)i) % N;
  i0 = (i1-1)   % N;
  i2 = (i1+1)   % N;
  i3 = (i1+2)   % N;
  return -y*(y-1)*(y-2)*T[i0]/6+(y+1)*(y-1)*(y-2)*T[i1]/2
    -(y+1)*y*(y-2)*T[i2]/2+(y+1)*y*(y-1)*T[i3]/6;
}


/*****************************************************************************
 * interpl
 ****************************************************************************/
__device__ t_float interpl(t_float *T, t_float i, int N)
{
  t_float y = i-(int)i;
  int i0,i1;

  i0 = ((int)i) % N;
  i1 = (i0+1)   % N;

  return (T[i1] - T[i0]) * y + T[i0];
}


/*****************************************************************************
 * modS
 ****************************************************************************/
__device__ t_float modS(t_float y, int S)
{
  return y-((int)y/S)*S;
}
          

extern "C" void cuda_pv_set_texture(t_cupv *x)
{
  //x->sineTextureFormatDesc = cudaCreateChannelDesc(32,0,0,0,cudaChannelFormatKindFloat);

  sineTexture.normalized = false;
  sineTexture.filterMode = cudaFilterModeLinear;
  sineTexture.addressMode[0] = cudaAddressModeClamp;
  sineTexture.channelDesc = x->sineTextureFormatDesc;

  CUDA_SAFE_CALL(cudaGetTextureReference((const textureReference **) &(x->sineTextureRef), "sineTexture"));

  CUDA_SAFE_CALL(cudaBindTextureToArray(sineTexture,
                                        x->sineTextureArray,
                                        x->sineTextureFormatDesc));
}
